package PureMessage::Scheduler::CronConvert;
use strict;
use Carp;
use ActiveState::Unix::Crontab qw(cron_parse cron_list cron_format);
use ActiveState::Config qw(:path);
use PureMessage::Scheduler::Config;

use base 'Exporter';

our @EXPORT_OK = qw(jobs2events);

# Converts a cron job to a scheduler job.
# Warning: times will be copied over even if the cron job is disabled!
#
# job_spec: hashref (or array of hashrefs) with fields:
#   id - optional
#   re - regexp with which to find the right cron job
#   disable_only
# jobs (optional)
# quiet (bool)

sub jobs2events {
    my %opts = @_;

    my $spec = delete $opts{job_spec}
        || croak "Missing required option 'job_spec'";
    $spec = ref($spec) eq "ARRAY" ? $spec : [ $spec ];

    my $jobs = delete $opts{jobs}
        || [ cron_list() ];

    # $jobs takes the form of an array of actual cron entries.

    my $quiet = delete $opts{quiet};

    if ($^W) {
	carp "Invalid option '$_'" foreach keys %opts;
    }

    for (@{$jobs}) {
	eval {
	    # %details has fields min, hour, mday, mon, wday, cmd, data
	    my %details = cron_parse($_);

	    foreach my $s (@{$spec}) {
		next unless $details{cmd} =~ $s->{re};

		# Use the cron job's time settings for the new sched
		# task.

		# Get the parse tree of all the scheduler tasks
                my $config = PureMessage::Scheduler::Config->new(writable => 1);
                my %tasks = $config->as_hash;

		# Pull out the corresponding scheduler task

		foreach my $id (keys %{$tasks{event}}) {
		    next unless $id =~ $s->{id};

		    # The copy we searched in was read-only - now get
		    # a writable version
		    my $e_node = path_node($config->root, "event#$id",
		    create => 'group') or carp "Failed to create config node";

		    # Store the cron times into the event node
		    path_value($e_node, 'when.s', 0);
		    path_value($e_node, 'when.m', $details{min});
		    path_value($e_node, 'when.h', $details{hour});
		    path_value($e_node, 'when.md', $details{mday});
		    path_value($e_node, 'when.mo', $details{mon});
		    path_value($e_node, 'when.wd', $details{wday});

		    $config->disable_event($id) if !$details{enabled};
		    $config->save;

		    last;
		}

		if ($details{enabled}) {
		    print "Disabling cron job '$details{cmd}'.\n" 
			unless $quiet;
		    $details{enabled} = 0;

		    # spit the line back out to crontab
		    $_ = cron_format(%details);
		}

		last;
	    }
	};
	carp "Failed to parse and modify cron job '$_': $@" if $@;
    }
    return @{$jobs};
}

1;

