package PureMessage::Scheduler::Event;
use strict;
use Time::Local qw(timelocal timelocal_nocheck timegm timegm_nocheck);
use Carp;
use base 'Exporter';

our @EXPORT_OK = qw(
    expand_time_range
    %TIME_RANGE
);

our %TIME_RANGE = (
    s  => [0..59],
    m  => [0..59],
    h  => [0..23],
    md => [1..31],
    mo => [0..11],
    wd => [0..6],
);

sub DAYS_TO_CACHE {6}; # MUST be > 1; 6 is ideal in most cases

# perldoc -f localtime
sub T_SECOND {0};
sub T_MINUTE {1};
sub T_HOUR   {2};
sub T_MDAY   {3};
sub T_MONTH  {4};
sub T_YEAR   {5};
sub T_WDAY   {6};
sub T_YDAY   {7};
sub T_DST    {8};
sub T_SMH    {(0..2)}
sub T_DMY    {(3..5)}

my %verbose_level = (
    "DEBUG"     => 3,
    "INFO"      => 2,
    "NOTICE"    => 1,
    "WARNING"   => 0,
    "ERR"       => 0,	# syslog compatible spelling
    "ERROR"     => 0,	# alias for ERR
    "CRIT"      => 0,
    "ALERT"     => 0,
    "EMERG"     => 0
);

sub new {
    my $class = shift;
    $class = ref($class) || $class;
    my %opts = @_;

    croak "Missing required parameter 'data'" unless $opts{data};
    croak "Missing required parameter 'id'"   unless $opts{id};

    my $self = bless \%opts, $class;
    $self->{data}{id} = $opts{id};

    $self->{handler} = eval {
	require PureMessage::Scheduler::Handler;
	PureMessage::Scheduler::Handler->get_handler(%{ $opts{data} },
                                                     logger       => $opts{logger},
                                                     notification => $opts{notification});
    };
    if ($@) {
	my $msg = "Failed to construct a handler for event $opts{id}: $@";
	$self->_log("WARNING", $msg);
	croak $msg;
    }

    $self->{init_time} ||= time;
    $self->compile($self->{init_time});

    return $self;
}

sub run {
    my $self = shift;
    $self->{handler}->run(%{ $self->{data} });
}

sub ready {
    my $self = shift;
    return $self->{handler}->ready;
}

sub _tsplit {
    # gsar says: "calling context propagates"
    my $self = shift;
    if ($self->{use_utc_time}) {
	return gmtime shift;
    } else {
	return localtime shift;
    }
}

sub _tjoin {
    my $self = shift;
    my $epoch;
    if ($self->{use_utc_time}) {
        $epoch = timegm(@_);
    } else {
        $epoch = timelocal(@_);
        
        # DEF29007 - Scheduler dies when time shifts from Daylight saving to
        # normal time When time changes from 1:59:00 to 1:00:00 in localtime,
        # timelocal will returns wrong epoch, because it doesn't know timezone.
        $epoch += 3600 if ($self->_in_duplicated_localtime());
    }
    return $epoch;
}

sub _tjoin_loose {
    my $self = shift;
    my $epoch;
    if ($self->{use_utc_time}) {
        $epoch = timegm_nocheck(@_);
    } else {
        $epoch = timelocal_nocheck(@_);

        # DEF29007 - Scheduler dies when time shifts from Daylight saving to
        # normal time When time changes from 1:59:00 to 1:00:00 in localtime,
        # timelocal will returns wrong epoch, because it doesn't know timezone.
        $epoch += 3600 if ($self->_in_duplicated_localtime());
    }
    return $epoch;
}

sub _in_duplicated_localtime {
    my $self = shift;
    my $now = time;

    my $after_convert = timelocal(localtime($now));
    return 0 if ($now == $after_convert);
    return 1;
}

sub expand_time_range {
    my ($t, $type) = @_;

    my $range = $TIME_RANGE{$type};
    my $t_min = $range->[0];
    my $t_max = $range->[$#$range];

    $t =~ s/\s*//g; 
    return [@$range] if ($t =~ /^\*$/); # copy of range
    my @l = split(',', $t);

    # look for ranges and repeats and convert them to lists
    my @l_final;
    foreach my $i (@l) {
	# range
	if ($i =~ /-/) {
	    my ($min, $max) = split('-', $i);
	    $min = $t_min if ($min < $t_min || $min > $t_max);
	    $max = $t_max if (defined $max && ($max > $t_max || $max < $t_min));
	    if (defined $max) {
		push(@l_final, ($min..$max));
	    }
	    else {
		push(@l_final, $min);
	    }
	} 
	# repeat
	elsif ($i =~ /\*\/(\d+)/) {
	    my $inc = $1 || 1;
	    for (my $j = $t_min; $j < $t_max; $j += $inc) {
		push(@l_final, $j);
	    }
	}
        elsif ($i eq '*') {
            push(@l_final, ($t_min .. $t_max));
        }
	# scalar
	else {
	    push(@l_final, $i);
	}
    }
    my %l_final_h = map {$_=>1} @l_final;
    @l_final = sort {$a<=>$b} (keys %l_final_h);
    return \@l_final;
}

sub _find_next_n_days {
    my $self = shift;
    my ($n,$now,$c_when) = @_;
    my @i_day = $self->_tsplit($now);
    @i_day[(T_SMH)] = (0,0,0);

    my $every_day = ($c_when->{star_md} && $c_when->{star_wd});
    my $md_wd_union = (!$c_when->{star_md} && !$c_when->{star_wd});
    my $ignore_wd = (!$c_when->{star_md} && $c_when->{star_wd});
    my $ignore_md = (!$c_when->{star_wd} && $c_when->{star_md});

    my %wd_map = map {$_ => 1} @{$c_when->{wd}} 
	unless ($every_day || $ignore_wd);
    my %md_map = map {$_ => 1} @{$c_when->{md}} 
	unless ($every_day || $ignore_md);
    my %mo_map = map {$_ => 1} @{$c_when->{mo}};
    
    # TODO : optimize this loop to skip over non-applicable months
    my @days;
    while (@days < $n) {
	if ($mo_map{$i_day[T_MONTH]} && 
	    (   $every_day ||
		($md_wd_union && 
		    ($md_map{$i_day[T_MDAY]} || $wd_map{$i_day[T_WDAY]})
		) ||
		($ignore_wd && $md_map{$i_day[T_MDAY]}) ||
		($ignore_md && $wd_map{$i_day[T_WDAY]})
	    )
	) {
	    push @days,[@i_day]; # copy of array
	}

	$i_day[3] += 1;
	@i_day = $self->_tsplit($self->_tjoin_loose(0,0,0,@i_day[(T_DMY)]));
    }

    return \@days;
}

sub _compile_event {
    my $self = shift;
    my $now = shift;
    die "must specify an when to compile the event"
	unless $now;
   
    my $w = $self->{data}{when};
    my %cw;
    
    foreach my $k (qw(s m h md wd mo)) {
        my $tr = defined($w->{$k}) ? $w->{$k} : '*';
	$cw{$k} = expand_time_range($tr, $k);
	$cw{"star_$k"} = 1 if ($tr eq '*');
    }

    $cw{yd} = $self->_find_next_n_days(DAYS_TO_CACHE,$now,\%cw);

    return \%cw;
}

sub compile {
    my $self = shift;
    my $now = shift || time;
    if ($self->{data}{when} && !$self->is_trigger_only) {
	$self->{compiled} = $self->_compile_event($now);
    }
    $self->calc_next_run($now);
    return 1;
}

sub calc_next_run {
    my ($self, $now) = @_;
    die "must specify current time to calc_next_run"
	unless $now;
    my $verb = ($self->{verbose} && 
	$self->{verbose} >= $verbose_level{'DEBUG'});

    return 0 if $self->is_trigger_only;

    # increment to the next second to deal with the 'all wildcard' case 
    $now++;
    $self->_log('DEBUG',"start: ".$self->_tsplit($now)) if $verb;

    # no 'when' block means run every second, every minute, ...
    unless (exists $self->{compiled}) {
	$self->{next_run} = $now;
	return;
    }
    my @now = $self->_tsplit($now);
    my $c = $self->{compiled};

    my ($now_s,$now_m,$now_h) = @now[(T_SMH)];
    my ($now_y,$now_yd) = @now[T_YEAR,T_YDAY];
    my $t_now_h = $self->_tjoin(0,0,@now[T_HOUR .. T_YEAR]);
    my $t_now_m = $self->_tjoin(0,@now[T_MINUTE .. T_YEAR]);
    my $t_now_s = $self->_tjoin(@now[T_SECOND .. T_YEAR]);

    # yd is an arrayref, rest are integer-scalars:
    my ($yd,$h,$m,$s) = map {$c->{$_}[0]} qw(yd h m s);

    my $round_up_days = sub {
	my $last = shift @{$c->{yd}};
	$yd = undef;
	while (!$yd) {
	    my @r_yd = grep { 
		($_->[T_YEAR] >= $now_y && $_->[T_YDAY] >= $now_yd) || 
		($_->[T_YEAR] > $now_y)
	    } @{$c->{yd}};

	    if (!@r_yd) {
		# big miss!!
		# make sure to get a fresh set of days:
		$self->_log('DEBUG','(getting more days)') if $verb;
		my $next = $self->_tjoin_loose(0,0,0,@$last[(T_DMY)]);
		$c->{yd} = $self->_find_next_n_days(DAYS_TO_CACHE,$next,$c);
		$last = $c->{yd}[$#{$c->{yd}}]; # so we don't loop forever
		next;
	    }
	    $c->{yd} = \@r_yd;
	    $yd = $r_yd[0];
	    $self->_log('DEBUG',
		'got day '.$self->_tsplit($self->_tjoin(@$yd))
	    ) if $verb;
	}
    };
    
    # We can incorrectly guess 4 successive approximations
    # (date, hour, minute, second) :
    my $safety = 4;
    my $success = 0;
    while ($safety--) {

	if ($verb) {
	    my $t = $self->_tsplit($self->_tjoin($s,$m,$h,@$yd[(T_DMY)]));
	    $self->_log('DEBUG','loop: '.$t);
	}
	
	if (!$yd || 
	    ($yd->[T_YEAR] < $now_y) ||
	    ($yd->[T_YEAR] == $now_y && $yd->[T_YDAY] < $now_yd)
	) {
	    # day has passed, reset hours, minutes, and seconds
	    $self->_log('DEBUG',"day in past (safe=$safety)") if $verb;
	    $h = $c->{h}[0]; 
	    $m = $c->{m}[0]; 
	    $s = $c->{s}[0];
	    &$round_up_days;
	    next;
	}
	
	if ($self->_tjoin(0,0,$h,@$yd[(T_DMY)]) < $t_now_h) {
	    # hour has passed, reset minutes and seconds
	    $self->_log('DEBUG',"hour mismatch (safe=$safety)") if $verb;
	    $m = $c->{m}[0]; 
	    $s = $c->{s}[0];
	    my @r_h = grep {$_ >= $now_h} @{$c->{h}};

	    if (!@r_h) { # day roll-over!
		$self->_log('DEBUG','..day roll-over') if $verb;
		&$round_up_days;
		next;
	    }
	    $h = $r_h[0];
	}

	if ($self->_tjoin(0,$m,$h,@$yd[(T_DMY)]) < $t_now_m) {
	    # minute has passed, reset seconds
	    $self->_log('DEBUG',"minute mismatch (safe=$safety)") if $verb;
	    $s = $c->{s}[0];
	    my @r_m = grep {$_ >= $now_m} @{$c->{m}};

	    if (!@r_m) { # hour roll-over! grep for hour
		$self->_log('DEBUG','..hour roll-over') if $verb;
		my @r_h = grep {$_ > $h} @{$c->{h}};
		if (!@r_h) { # day roll-over!
		    $self->_log('DEBUG','...day roll-over') if $verb;
		    $h = $c->{h}[0];
		    &$round_up_days;
		} 
		else {
		    $h = $r_h[0];
		}
		next;
	    }
	    $m = $r_m[0];
	}

	if ($self->_tjoin($s,$m,$h,@$yd[(T_DMY)]) < $t_now_s) {
	    # second has passed, look for a good one
	    $self->_log('DEBUG',"second mismatch (safe=$safety)") if $verb;
	    my @r_s = grep {$_ >= $now_s} @{$c->{s}};

	    if (!@r_s) { # minute roll-over! grep for minute
		$self->_log('DEBUG','..minute roll-over') if $verb;
		my @r_m = grep {$_ > $m} @{$c->{m}};

		if (!@r_m) { # hour roll-over! grep for hour
		    $self->_log('DEBUG','...hour roll-over') if $verb;
		    $m = $c->{m}[0];
		    my @r_h = grep {$_ > $h} @{$c->{h}};

		    if (!@r_h) { # day roll-over!
			$self->_log('DEBUG','....day roll-over') if $verb;
			$h = $c->{h}[0];
			&$round_up_days;
		    } 
		    else {
			$h = $r_h[0];
		    }
		}
		else {
		    $m = $r_m[0];
		}
		next;
	    }
	    $s = $r_s[0];
	}

	$success = 1;
	last; #!
    }

    die "unable to calculate next run time for $self->{data}{id}\n"
	unless $success;

    $self->{next_run} = $self->_tjoin($s,$m,$h,@$yd[(T_DMY)]);
    if ($verb) {
	my $nr = $self->_tsplit($self->{next_run});
	$self->_log('DEBUG','next run: '.$nr);
    }
    return $self->{next_run};
}

sub is_trigger_only {
    return $_[0]->{data}{trigger_only};
}

sub next_run {
    return $_[0]->{next_run};
}

sub _dump_compiled_times{
    my ($self, $id, $ce) = @_;
    # TODO: make yd output meaningful
    $self->_log('DEBUG',"Compiled Times for event $id:");
    foreach my $k (keys %$ce) {
	$self->_log('DEBUG',".  $k: ".join(", ", @{$$ce{$k}}));
    }
}

sub _log {
    my ($self, $type, $msg) = @_;
    $self->{logger}->log($type, $msg) if $self->{logger};
    print STDERR "$type: $msg\n"
        if $self->{verbose} && $self->{verbose} >= $verbose_level{$type};
}

1;
