package PureMessage::Scheduler::Handler::exec;
use strict;
use Carp;
use POSIX qw(:sys_wait_h);
use PureMessage::Config qw(%config);

sub new {
    my $class = shift;
    return bless { @_ }, $class;
}

sub run {
    my $self = shift;
    my %data = @_;

    for (qw(action id)) {
	croak "Missing required parameter '$_'" unless length $data{$_};
    }

    $self->_log("DEBUG",
	"Run \"$data{action}\" for $data{id}");

    my $notifier = $self->{notification};
    my $pid = fork();
    if (defined $pid) {
        if ($pid) {
            $self->{wait_pid} = $pid;
            $notifier->event_info($pid, data => \%data) if $notifier;
            $self->_log("DEBUG", "Exec child pid [$pid]");
            return $pid;
        }
        else {
            # We *know* we probably won't reach the code after exec().
            no warnings 'exec';

            if ($notifier) {
                my $job_logfile = $notifier->job_filename();
                if (open(FH, ">$job_logfile")) {
                    open(STDOUT, ">&=FH");
                    open(STDERR, ">&=FH");
                }
                else {
                    $self->_log("ERROR", "Failed to write job logfile [$job_logfile]: $!");
                }
            }

            exec($data{action});
            eval {
                $self->_log("ERROR", "Failed to run \"$data{action}\": $!");
            };

            # make sure that we call POSIX::_exit() so that we don't end up calling
            # DESTROY methods from this process.
            POSIX::_exit(1); # ensure we never return to parent code
        }
    }
    else {
        # Error - probably out of memory 
        $self->_log("ERROR", "Failed to fork to run \"$data{action}\": $!");
        $self->{wait_pid} = undef;
        return undef;
    }
}

sub ready {
    my $self = shift;
    return 1 unless $self->{wait_pid};

    my $rc = waitpid($self->{wait_pid}, WNOHANG);
    if ($rc == -1) {
        $self->_log("ERROR", "Error occured while waiting for child [$self->{wait_pid}]: $!");
        $self->{wait_pid} = undef;
        return 1;
    }
    elsif ($rc == $self->{wait_pid}) {
        my $mesg;
        my %status = (
            exit_code => $? >> 8,
            signal    => $? & 127,
            coredump  => $? & 128,
        );
        $mesg  = "Exit code: " . $status{exit_code};
        $mesg .= ", Signal: $status{signal}" if $status{signal};

	$self->_log("DEBUG", "Child exited [$self->{wait_pid}] - $mesg");
	$self->_log("DEBUG", "Handler is ready");

        eval {
            my $notifier;
            if ($notifier = $self->{notification}) {
                $notifier->notify($self->{wait_pid}, status => \%status);
            }
        };
        $self->_log("ERROR", "Notification failed") if $@;
        $self->{wait_pid} = undef;

	return 1;
    }
        
    # suppress this unless we really want verbosity.    
    if ($config{debug_level} >= 6) {
        $self->_log("DEBUG", "Handler is still busy") unless $ENV{PMX_SCHED_SANE_LOGGING};
    }
        
    return 0;
}

sub _log {
    my ($self, $type, $msg) = @_;
    $self->{logger}->log($type, $msg) if $self->{logger};
}

1;
