package PureMessage::Scheduler::Notify;
use strict;
use Carp;
use HTTP::Date qw(time2iso);
use PureMessage::Config qw(%config);
use ActiveState::Handy qw(cat);
use MIME::Entity;
use File::stat;

my $JOB_DIR         = $config{run_dir} . "/scheduler";
my $JOB_LOGFILENAME = "scheduler.out";

sub new {
    my $class = shift;
    $class = ref($class) || $class;
    my %opts = @_;

    croak "Missing required parameter 'from_addr'" unless $opts{from_addr};
    croak "Missing required parameter 'to_addr'"   unless $opts{to_addr};

    my $self = bless \%opts, $class;
    $self->{event_info} = {};

    # Assign full path to logfile
    $self->_event_logfile($opts{event_logfile}) if $opts{event_logfile};
    $self->_remove_stale_logfiles();
    return $self;
}

sub notify {
    my ($self, $id, %opts) = @_;
    my $trigger_data;

    # Return and remove event information
    my $ev = $self->event_info($id, purge => 1);

    # Retrieve output
    my $job_log     = $self->job_filename($id);
    my $job_output  = cat($job_log) || '';
    my $status      = $opts{status};
    unlink($job_log);

    return unless ($status->{exit_code} || $self->event_logging_enabled);

    # Error?
    my $mesg = $self->build_notification($id,
                                         event  => $ev,
                                         status => $status,
                                         output => $job_output);
    return unless $mesg;

    if ($status->{exit_code}) {
        # Log errors
        $self->_event_log($mesg);

        if ($self->notify_on_error) {
            $self->queue_notification($id,
                                      event  => $ev,
                                      output => $job_output);
        }
    }
    elsif ($self->event_logging_enabled) {
        # XXX Log all
        $self->_event_log($mesg);
    }
    return 1;
}

sub build_notification {
    my ($self, $id, %opts) = @_;

    unless ($id) {
        $self->_log("ERROR", "Cant notify without a job id");
        return;
    }

    my $ev_id      = $opts{event}  ? $opts{event}->{id}         : 'unknown';
    my $action     = $opts{event}  ? $opts{event}->{action}     : 'unknown';
    my $exit_code  = $opts{status} ? $opts{status}->{exit_code} : 'unknown';
    my $job_output = $opts{output} || '';

    my $now       = time();
    my $data      = join("\n", 
                         "Time      : " . time2iso($now),
                         "PID       : " . $id,
                         "Job ID    : " . $ev_id,
                         "Command   : " . $action,
                         "Exit code : " . $exit_code,
                         "\n$job_output",
                        );
    return $data;
}

sub queue_notification {
    my ($self, $id, %params) = @_;

    croak "Missing required parameter 'params'" unless %params;
    croak "Missing required parameter 'output'" unless exists $params{output};

    my $to         = $self->{to_addr};
    my $from       = $self->{from_addr};
    my $ev_id      = ($params{event}) ? $params{event}->{id} : 'unknown';
    my $message    = $params{output} || '';

    my $mimemsg = MIME::Entity->build(
         Subject => "Scheduler job notification [error] [$ev_id]",
         To      => $to,
         From    => $from,
         Type    => 'text/plain',
         Data    => $message,
    );

    eval {
        require PureMessage::Queue;
        PureMessage::Queue::queue_mail(
            'to'      => $to,
            'from'    => $from,
            'message' => $mimemsg->stringify(),
        );
    };
    if ($@) {
        $self->_log("ERROR", "Unable to queue failure notification: $@");
        return;
    }
    return 1;
}

sub event_info {
    my ($self, $id, %opts) = @_;
    my $ev;

    unless ($id && !ref($id)) {
        $self->_log("ERROR", "Event info request without job id");
        return;
    }

    if ($opts{data}) {
        $ev = $self->{event_info}->{$id} = $opts{data};
    }
    else {
        $ev = $self->{event_info}->{$id};
        delete $self->{event_info}->{$id} if ($opts{purge});
    }
    return $ev;
}

sub job_filename {
    my ($self, $id) = @_;
    $id ||= $$;
    return "$JOB_DIR/$JOB_LOGFILENAME.$id";
}

sub notify_on_error {
    my $self = shift;
    if (@_) {
        $self->{notify_on_error} = shift;
    }
    return $self->{notify_on_error};    
}

sub event_logging_enabled {
    my $self = shift;
    $self->{event_logging} = shift if @_;
    return $self->{event_logging};    
}

sub _remove_stale_logfiles {
    my ($self, %opts) = @_;
    my $stale_time = $opts{stale_time} || 14400; # 4h
    my $now = time();

    if (-d $JOB_DIR) {
        my $rem_count = 0;
        foreach my $file (<$JOB_DIR/$JOB_LOGFILENAME.*>) {
            my $sb = stat($file);
            if ($sb && (($now - $sb->mtime) >= $stale_time)) {
                unlink($file) ? $rem_count++ : warn("Unable to remove $file: [$!]\n");
            }
        }
        $self->_log("INFO", "Removed $rem_count stale job log(s)") if $rem_count;
    }
}

sub _log {
    my ($self, $type, $msg) = @_;
    $self->{logger}->log($type, $msg) if $self->{logger};
}

sub _event_logfile {
    my $self = shift;
    if (@_) {
        my $filename = shift;
        if ($filename =~ /^\//) {
            $self->{event_logfile} = $filename;
        }
        else {
            $self->{event_logfile} = $config{log_dir} . "/" . $filename;
        }
    }
    return $self->{event_logfile};
}

sub _event_log {
    my ($self, $msg) = @_;
    my $filename = $self->_event_logfile();
    if ($filename) {
        if (open(my $fh, ">>$filename")) {
            chomp($msg);
            print $fh "$msg\n" . "-" x 31 . "\n";
            close $fh;
        }
        else {
            $self->_log("ERROR", "Cant open log '$filename' [$!]");
        }
    }
}
1;
